#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdarg.h>

#include "libcr.h"

sig_atomic_t child_pid = 0; /* who did we spawn? */
static void signal_child (int, siginfo_t *, void *);
static int verbose = 1;

static void warn(const char *format, ...)
                __attribute__ ((format (printf, 1, 2)));
/*
 * warn(format, args...)
 *
 * Sends a warning message to stderr
 */
static void warn(const char *format, ...)
{
    va_list args;

    va_start(args, format);
    if (verbose >= 0) {
      vfprintf(stderr, format, args);
    }
    va_end(args);
}

static void signal_self(int sig)
{
    struct sigaction sa;

    /* restore default (in kernel) handler */
    sa.sa_handler = SIG_DFL;
    sa.sa_flags = SA_RESTART | SA_NOMASK;
    sigemptyset(&sa.sa_mask);
    (void)sigaction(sig, &sa, NULL);

    /* send to self */
    raise(sig);

    /* restore self as handler */
    sa.sa_sigaction = &signal_child;
    sa.sa_flags = SA_RESTART | SA_NOMASK | SA_SIGINFO;
    (void)sigaction(sig, &sa, NULL);
}

/* Best-effort to forward all signals to the restarted child
 * XXX: This is probably not complete in terms of special-cases
 */
static void signal_child (int sig, siginfo_t *siginfo, void *context)
{
    if ((siginfo->si_code > 0) &&       /* si_code > 0 indicates sent by kernel */
        (sig == SIGILL || sig == SIGFPE || sig == SIGBUS || sig == SIGSEGV )) {
        /* This signal is OUR error, so we don't forward */
        signal_self(sig);
    } else if (sig == SIGTSTP || sig == SIGTTIN || sig == SIGTTOU) {
        /* The catchable stop signals go to child AND self */
        (void)kill(child_pid, sig);
        signal_self(sig);
    } else {
        /* default case */
        (void)kill(child_pid, sig);
    }
}

/* try to exit such that our exit code is same as our child */
void mimic_exit(int status)
{
    if (WIFEXITED(status)) {
        /* easy to mimic normal return */
        exit(WEXITSTATUS(status));
    } else if (WIFSIGNALED(status)) {
        /* disable generation of a 'core' */
        struct rlimit r;
        r.rlim_cur = r.rlim_max = 0;
        (void)setrlimit(RLIMIT_CORE, &r);

        /* now raise the signal */
        signal_self(WTERMSIG(status));
    } else {
        warn("Unexpected status from child\n");
        exit(-1);
    }
}


void init_signals(sigset_t *oldset)
{
    sigset_t set;
    struct sigaction sa;
    int sig;

    sigfillset(&set);
    sigdelset(&set, SIGSTOP);
    sigdelset(&set, SIGKILL);
    sigdelset(&set, SIGCHLD);
    sigprocmask(SIG_SETMASK, &set, oldset);
    sa.sa_sigaction = &signal_child;
    sa.sa_flags = SA_RESTART | SA_NOMASK | SA_SIGINFO;
    sigemptyset(&sa.sa_mask);
    for (sig = 1; sig < _NSIG; ++sig) {
        if (sig == SIGCHLD) continue;
        if (sig == SIGKILL) continue;
        if (sig == SIGSTOP) continue;
        if (sig == CR_SIGNUM) continue;

        /* best-effort so we ignore failures */
        (void)sigaction(sig, &sa, NULL);
    }
}

void reset_signals(const sigset_t *set)
{
    int sig;
    struct sigaction sa;

    sa.sa_handler = SIG_DFL;
    sigemptyset(&sa.sa_mask);

    for (sig = 1; sig < _NSIG; ++sig) {
        if (sig == SIGCHLD) continue;
        if (sig == SIGKILL) continue;
        if (sig == SIGSTOP) continue;
        if (sig == CR_SIGNUM) continue;

        /* best-effort so we ignore failures */
        (void)sigaction(sig, &sa, NULL);
    }

    sigprocmask(SIG_SETMASK, set, NULL);
}

void forward_signals(const sigset_t *set, int pid)
{
    child_pid = pid;
    sigprocmask(SIG_SETMASK, set, NULL);
}
